import aniso8601
from datetime import datetime
from language_tags import tags as language_tags
import numbers
from pyld import jsonld
from pytz import utc
import re
import six

from ..actions.graph import patch_node
from ..actions.tasks import add_task, report_message
from ..exceptions import TaskPrerequisitesError, ValidationError
from ..state import get_node_by_id, get_node_by_path
from ..openbadges_context import OPENBADGES_CONTEXT_V2_DICT
from ..utils import list_of, MESSAGE_LEVEL_WARNING

from .task_types import (ASSERTION_TIMESTAMP_CHECKS, ASSERTION_VERIFICATION_CHECK,
                         ASSERTION_VERIFICATION_DEPENDENCIES, CLASS_VALIDATION_TASKS,
                         CRITERIA_PROPERTY_DEPENDENCIES, FETCH_HTTP_NODE, FLATTEN_EMBEDDED_RESOURCE,
                         HOSTED_ID_IN_VERIFICATION_SCOPE, IDENTITY_OBJECT_PROPERTY_DEPENDENCIES,
                         IMAGE_VALIDATION, ISSUER_PROPERTY_DEPENDENCIES, VALIDATE_EXPECTED_NODE_CLASS,
                         VALIDATE_RDF_TYPE_PROPERTY, VALIDATE_PROPERTY, VALIDATE_REVOCATIONLIST_ENTRIES,
                         VERIFY_RECIPIENT_IDENTIFIER)
from .utils import (abbreviate_value as abv,
                    abbreviate_node_id as abv_node,
                    is_blank_node_id, is_data_uri, is_empty_list, is_null_list, is_iri, is_url, task_result, )

from future.standard_library import install_aliases

install_aliases()
from urllib.parse import urlparse


class OBClasses(object):
    AlignmentObject = 'AlignmentObject'
    Assertion = 'Assertion'
    BadgeClass = 'BadgeClass'
    Criteria = 'Criteria'
    CryptographicKey = 'CryptographicKey'
    Endorsement = 'Endorsement'
    EndorsementClaim = 'EndorsementClaim'
    Evidence = 'Evidence'
    ExpectedRecipientProfile = 'ExpectedRecipientProfile'
    Extension = 'Extension'
    IdentityObject = 'IdentityObject'
    Image = 'Image'
    Issuer = 'Issuer'
    Profile = 'Profile'
    RevocationList = 'RevocationList'
    VerificationObject = 'VerificationObject'

    VerificationObjectAssertion = 'VerificationObjectAssertion'
    VerificationObjectIssuer = 'VerificationObjectIssuer'

    ALL_CLASSES = (AlignmentObject, Assertion, BadgeClass, Criteria, CryptographicKey, Endorsement, EndorsementClaim,
                   ExpectedRecipientProfile, Extension, Evidence, IdentityObject, Image, Issuer,
                   Profile, RevocationList, VerificationObject, VerificationObjectAssertion,
                   VerificationObjectIssuer)
    PRIMARY_OBJECTS = ([Assertion, BadgeClass, Issuer, Profile, Endorsement])

    @classmethod
    def default_for(cls, class_name):
        defaults = {
            cls.Issuer: cls.Profile
        }

        if class_name not in cls.ALL_CLASSES:
            raise TypeError("Unknown class name {}".format(class_name))
        try:
            return defaults[class_name]
        except KeyError:
            return class_name


class ValueTypes(object):
    BOOLEAN = 'BOOLEAN'
    COMPACT_IRI = 'COMPACT_IRI'
    DATA_URI = 'DATA_URI'
    DATA_URI_OR_URL = 'DATA_URI_OR_URL'
    DATETIME = 'DATETIME'
    EMAIL = 'EMAIL'
    ID = 'ID'
    IDENTITY_HASH = 'IDENTITY_HASH'
    IRI = 'IRI'
    LANGUAGE = 'LANGUAGE'
    MARKDOWN_TEXT = 'MARKDOWN_TEXT'
    RDF_TYPE = 'RDF_TYPE'
    TELEPHONE = 'TELEPHONE'
    TEXT = 'TEXT'
    TEXT_OR_NUMBER = 'TEXT_OR_NUMBER'
    URL = 'URL'
    URL_AUTHORITY = 'URL_AUTHORITY'

    PRIMITIVES = (
        BOOLEAN, DATA_URI_OR_URL, DATETIME, ID, IDENTITY_HASH, IRI, LANGUAGE, MARKDOWN_TEXT,
        TELEPHONE, TEXT, TEXT_OR_NUMBER, URL, URL_AUTHORITY
    )


class PrimitiveValueValidator(object):
    """
    A callable validator for primitive Open Badges value types. 
    
    Example usage: 
    PrimitiveValueValidator(ValueTypes.TEXT)("test value")
    > True
    """

    def __init__(self, value_type):
        value_check_functions = {
            ValueTypes.BOOLEAN: self._validate_boolean,
            ValueTypes.DATA_URI: self._validate_data_uri,
            ValueTypes.DATA_URI_OR_URL: self._validate_data_uri_or_url,
            ValueTypes.DATETIME: self._validate_datetime,
            ValueTypes.EMAIL: self._validate_email,
            ValueTypes.IDENTITY_HASH: self._validate_identity_hash,
            ValueTypes.IRI: self._validate_iri,
            ValueTypes.LANGUAGE: self._validate_language,
            ValueTypes.COMPACT_IRI: self._validate_compact_iri,
            ValueTypes.MARKDOWN_TEXT: self._validate_markdown_text,
            ValueTypes.RDF_TYPE: self._validate_rdf_type,
            ValueTypes.TELEPHONE: self._validate_tel,
            ValueTypes.TEXT: self._validate_text,
            ValueTypes.TEXT_OR_NUMBER: self._validate_text_or_number,
            ValueTypes.URL: self._validate_url,
            ValueTypes.URL_AUTHORITY: self._validate_url_authority
        }
        self.value_type = value_type
        self.is_valid = value_check_functions[value_type]

    def __call__(self, value):
        return self.is_valid(value)

    @staticmethod
    def _validate_boolean(value):
        return isinstance(value, bool)

    @classmethod
    def _validate_compact_iri(cls, value):
        if value == 'id' or cls._validate_iri(value):
            return True
        try:
            test_data = {'@context': OPENBADGES_CONTEXT_V2_DICT, value: 'TEST'}
            expanded = jsonld.expand(test_data)
            if len(list(expanded[0])) == 1:
                return True
        except (jsonld.JsonLdError, IndexError):
            pass
        return False

    @staticmethod
    def _validate_data_uri(value):
        return is_data_uri(value)

    @classmethod
    def _validate_data_uri_or_url(cls, value):
        return bool(cls._validate_url(value) or cls._validate_data_uri(value))

    @staticmethod
    def _validate_datetime(value):
        try:
            # aniso at least needs to think it can get a datetime from value
            aniso8601.parse_datetime(value)
        except Exception as e:
            return False
        # we also require tzinfo specification on our datetime strings
        # NOTE -- does not catch minus-sign (non-ascii char) tzinfo delimiter
        return (isinstance(value, six.string_types) and
                (value[-1:] == 'Z' or
                 bool(re.match(r'.*[+-](?:\d{4}|\d{2}|\d{2}:\d{2})$', value))))

    @staticmethod
    def _validate_email(value):
        return bool(re.match(r'(^[^@\s]+@[^@\s]+$)', value))

    @staticmethod
    def is_hashed_identity_hash(value):
        return bool(re.match(r'md5\$[\da-fA-F]{32}$', value) or re.match(r'sha256\$[\da-fA-F]{64}$', value))

    @classmethod
    def _validate_identity_hash(cls, value):
        # Validates that identity is a string. More specific rules may only be enforced at the class instance level.
        return isinstance(value, six.string_types)

    @classmethod
    def _validate_iri(cls, value):
        """
        Checks if a string matches an acceptable IRI format and scheme. For now, only accepts a few schemes,
        'http', 'https', blank node identifiers, and 'urn:uuid'
        :param value: six.string_types 
        :return: bool
        """
        return is_iri(value)

    @staticmethod
    def _validate_language(value):
        return isinstance(value, six.string_types) and language_tags.check(value)

    @classmethod
    def _validate_markdown_text(cls, value):
        # All markdown is "valid" if it is a string.
        return cls._validate_text

    @classmethod
    def _validate_rdf_type(cls, value):
        try:
            if not (isinstance(value, six.string_types)):
                raise ValidationError(
                    'RDF_TYPE entry {} must be a string value'.format(abv(value)))

            expanded = jsonld.expand({"@context": OPENBADGES_CONTEXT_V2_DICT, 'type': value})
            expanded_value = expanded[0]['@type'][0]
            if not cls._validate_iri(expanded_value):
                raise ValidationError(
                    'RDF_TYPE entry {} must be a valid IRI in the document context'.format(
                        abv(value))
                )
        except (ValidationError, jsonld.JsonLdError,):
            return False

        return True

    @staticmethod
    def _validate_tel(value):
        """ Validates whether item passes E.164 validation (allows extensions)"""
        return bool(re.match(r'^\+?[1-9]\d{1,14}(;ext=\d+)?$', value))

    @staticmethod
    def _validate_text(value):
        return isinstance(value, six.string_types)

    @classmethod
    def _validate_text_or_number(cls, value):
        return cls._validate_text(value) or isinstance(value, numbers.Number)

    @staticmethod
    def _validate_url(value):
        return is_url(value)

    @staticmethod
    def _validate_url_authority(value):
        if not isinstance(value, six.string_types):
            return False

        hostname = re.compile(r'(?=^.{4,253}$)(^((?!-)[a-zA-Z0-9-]{0,62}[a-zA-Z0-9]\.)+[a-zA-Z]{2,63}$)')

        test_value = 'http://{}/test'.format(value)
        parsed = urlparse(test_value)
        au = parsed.netloc
        if not au or not hostname.match(au):
            return False

        return parsed.scheme == 'http' and au == value and parsed.path == '/test' and not parsed.query


def validate_property(state, task_meta, **options):
    """
    Validates presence and data type of a single property that is
    expected to be one of the Open Badges Primitive data types or an ID.
    """
    node_id = task_meta.get('node_id')
    node_path = task_meta.get('node_path')
    if node_id:
        node = get_node_by_id(state, node_id)
    elif node_path:
        node = get_node_by_path(state, node_path)
    node_class = task_meta.get('node_class', 'unknown type node')

    prop_name = task_meta.get('prop_name')
    prop_type = task_meta.get('prop_type')
    required = bool(task_meta.get('required'))
    allow_many = task_meta.get('many')
    depth = task_meta.get('depth')

    actions = []

    try:
        prop_value = node[prop_name]
    except KeyError:
        if not required:
            return task_result(
                True, "Optional property {} not present in {} {}".format(
                    prop_name, node_class, abv_node(node_id, node_path))
            )
        return task_result(
            False, "Required property {} not present in {} {}".format(
                prop_name, node_class, abv_node(node_id, node_path))
        )

    if not isinstance(prop_value, (list, tuple)):
        values_to_test = [prop_value]
    else:
        values_to_test = prop_value

    if required and (is_empty_list(values_to_test) or is_null_list(values_to_test)):
        return task_result(
            False, "Required property {} value {} is not acceptable in {} {}".format(
                prop_name, abv(prop_value), node_class, abv_node(node_id, node_path))
        )
    if not required and (is_empty_list(values_to_test) or is_null_list(values_to_test)):
        return task_result(True, "Optional property {} is null in {} {}".format(
            prop_name, node_class, abv_node(node_id, node_path)
        ))

    if not allow_many and len(values_to_test) > 1:
        return task_result(
            False, "Property {} in {} {} has more than the single allowed value.".format(
                prop_name, node_class, abv_node(node_id, node_path)
            ))

    try:
        if prop_type != ValueTypes.ID:
            for val in values_to_test:
                value_check_function = PrimitiveValueValidator(prop_type)
                if not value_check_function(val):
                    raise ValidationError("{} property {} value {} not valid in {} {}".format(
                        prop_type, prop_name, abv(val), node_class, abv_node(node_id, node_path)))
        else:
            for i in range(len(values_to_test)):
                val = values_to_test[i]
                if isinstance(prop_value, (list, tuple,)):
                    value_to_test_path = [node_id, prop_name, i]
                else:
                    value_to_test_path = [node_id, prop_name]

                if isinstance(val, dict):
                    actions.append(
                        add_task(VALIDATE_EXPECTED_NODE_CLASS, node_path=value_to_test_path,
                                 expected_class=task_meta.get('expected_class'),
                                 full_validate=task_meta.get('full_validate', True),
                                 depth=depth
                                 )
                    )
                    continue
                elif task_meta.get('allow_data_uri') and not PrimitiveValueValidator(ValueTypes.DATA_URI_OR_URL)(val):
                    raise ValidationError("ID-type property {} had value `{}` that isn't URI or DATA URI in {}.".format(
                        prop_name, abv(val), abv_node(node_id, node_path))
                    )
                elif not task_meta.get('allow_data_uri', False) and not PrimitiveValueValidator(ValueTypes.IRI)(val):
                    actions.append(report_message(
                        "ID-type property {} had value `{}` where another scheme may have been expected {}.".format(
                            prop_name, abv(val), abv_node(node_id, node_path)
                        ), message_level=MESSAGE_LEVEL_WARNING))
                    raise ValidationError(
                        "ID-type property {} had value `{}` not embedded node or in IRI format in {}.".format(
                            prop_name, abv(val), abv_node(node_id, node_path))
                    )
                try:
                    target = get_node_by_id(state, val)
                except IndexError:
                    if not task_meta.get('fetch', False):
                        if task_meta.get('allow_remote_url') and PrimitiveValueValidator(ValueTypes.URL)(val):
                            continue
                        if task_meta.get('allow_data_uri') and PrimitiveValueValidator(ValueTypes.DATA_URI)(val):
                            continue
                        raise ValidationError(
                            'Node {} has {} property value `{}` that appears not to be in URI format'.format(
                                abv_node(node_id, node_path), prop_name, abv(val)
                            ) + ' or did not correspond to a known local node.')
                    else:
                        actions.append(
                            add_task(FETCH_HTTP_NODE, url=val,
                                     expected_class=task_meta.get('expected_class'),
                                     source_node_path=value_to_test_path,
                                     depth=depth
                                     ))
                else:
                    actions.append(
                        add_task(VALIDATE_EXPECTED_NODE_CLASS, node_id=val,
                                 expected_class=task_meta.get('expected_class'),
                                 depth=depth
                                 )
                    )

    except ValidationError as e:
        return task_result(False, e.message)
    return task_result(
        True, "{} property {} is valid in {} {}".format(
            prop_type, prop_name, node_class, abv_node(node_id, node_path)
        ), actions
    )


def validate_rdf_type_property(state, task_meta, **options):
    prop_result = validate_property(state, task_meta)
    if not prop_result[0]:
        return prop_result

    node_id = task_meta.get('node_id')
    node_path = task_meta.get('node_path')
    if node_id:
        node = get_node_by_id(state, node_id)
    elif node_path:
        node = get_node_by_path(state, node_path)
    prop_value = node.get('type')
    required = bool(task_meta.get('required'))
    default = task_meta.get('default')
    must_contain_one = task_meta.get('must_contain_one')
    allow_many = task_meta.get('many', False)
    actions = []

    # Set node type to default value
    if not prop_value and default:
        actions.append(patch_node(node_id, {'type': default}))
        return task_result(True, prop_result[1], actions)

    if not isinstance(prop_value, (list, tuple)):
        values_to_test = [prop_value]
    else:
        values_to_test = prop_value

    # Reject if value not in allowed set of values.
    if must_contain_one and not any(val in must_contain_one for val in values_to_test):
        return task_result(False, 'Node {} of type {} does not have type among allowed values ({})'.format(
            abv_node(node_id, node_path), abv(prop_value), abv(must_contain_one)))

    return prop_result


class ClassValidators(OBClasses):
    def __init__(self, class_name):
        self.class_name = class_name

        if class_name == OBClasses.Assertion:
            self.validators = [
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': True},
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': True,
                 'many': True, 'must_contain_one': ['Assertion']},
                {'prop_name': 'recipient', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.IdentityObject, 'required': True},
                {'prop_name': 'badge', 'prop_type': ValueTypes.ID, 'prerequisites': 'ASN_FLATTEN_BC',
                 'expected_class': OBClasses.BadgeClass, 'fetch': True, 'required': True},
                {'prop_name': 'verification', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.VerificationObjectAssertion, 'required': True},
                {'prop_name': 'issuedOn', 'prop_type': ValueTypes.DATETIME, 'required': True},
                {'prop_name': 'expires', 'prop_type': ValueTypes.DATETIME, 'required': False},
                {'prop_name': 'image', 'prop_type': ValueTypes.ID, 'required': False, 'allow_remote_url': True,
                 'expected_class': OBClasses.Image, 'fetch': False, 'allow_data_uri': False},
                {'prop_name': 'narrative', 'prop_type': ValueTypes.MARKDOWN_TEXT, 'required': False},
                {'prop_name': 'evidence', 'prop_type': ValueTypes.ID, 'allow_remote_url': True,
                 'expected_class': OBClasses.Evidence, 'many': True, 'fetch': False, 'required': False},
                {'task_type': ASSERTION_VERIFICATION_DEPENDENCIES, 'prerequisites': ISSUER_PROPERTY_DEPENDENCIES},
                {'task_type': ASSERTION_TIMESTAMP_CHECKS},
                {'task_type': IMAGE_VALIDATION, 'prop_name': 'image', 'node_class': OBClasses.Assertion,
                 'required': False, 'many': False, 'allow_data_uri': False},
                {'task_type': FLATTEN_EMBEDDED_RESOURCE, 'prop_name': 'badge', 'node_class': OBClasses.Assertion,
                 'task_key': 'ASN_FLATTEN_BC'}
            ]
        elif class_name == OBClasses.BadgeClass:
            self.validators = [
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': True},
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': True,
                 'many': True, 'must_contain_one': ['BadgeClass']},
                {'prop_name': 'issuer', 'prop_type': ValueTypes.ID, 'prerequisites': 'BC_FLATTEN_ISS',
                 'expected_class': OBClasses.Profile, 'fetch': True, 'required': True},
                {'prop_name': 'name', 'prop_type': ValueTypes.TEXT, 'required': True},
                {'prop_name': 'description', 'prop_type': ValueTypes.TEXT, 'required': True},
                {'prop_name': 'image', 'prop_type': ValueTypes.ID, 'required': False, 'allow_remote_url': True,
                 'expected_class': OBClasses.Image, 'fetch': False, 'allow_data_uri': True},
                {'prop_name': 'criteria', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.Criteria, 'fetch': False,
                 'required': True, 'allow_remote_url': True},
                {'prop_name': 'alignment', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.AlignmentObject, 'many': True, 'fetch': False, 'required': False},
                {'prop_name': 'tags', 'prop_type': ValueTypes.TEXT, 'many': True, 'required': False},
                {'task_type': IMAGE_VALIDATION, 'prop_name': 'image', 'node_class': OBClasses.BadgeClass,
                 'required': True, 'many': False, 'allow_data_uri': True},
                {'task_type': FLATTEN_EMBEDDED_RESOURCE, 'prop_name': 'issuer', 'node_class': OBClasses.BadgeClass,
                 'task_key': 'BC_FLATTEN_ISS'}
            ]
        elif class_name == OBClasses.CryptographicKey:
            self.validators = [
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': False},
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': False, 'many': True,
                 'default': 'CryptographicKey'},
                {'prop_name': 'owner', 'prop_type': ValueTypes.IRI, 'required': False, 'fetch': True},
                {'prop_name': 'publicKeyPem', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'task_type': ASSERTION_VERIFICATION_CHECK, 'triggers_completion': 'SIGNING_KEY_FETCHED'}
            ]
        elif class_name == OBClasses.Profile or class_name == OBClasses.Issuer:
            # To start, required values will assume the Profile class is used as BadgeClass.issuer
            self.validators = [
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': True},
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': True,
                 'many': True, 'must_contain_one': ['Issuer', 'Profile']},
                {'prop_name': 'name', 'prop_type': ValueTypes.TEXT, 'required': True},
                {'prop_name': 'description', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'image', 'prop_type': ValueTypes.ID, 'required': False, 'allow_remote_url': True,
                 'expected_class': OBClasses.Image, 'fetch': False, 'allow_data_uri': True},
                {'prop_name': 'url', 'prop_type': ValueTypes.URL, 'required': True},
                {'prop_name': 'email', 'prop_type': ValueTypes.EMAIL, 'required': True},
                {'prop_name': 'telephone', 'prop_type': ValueTypes.TELEPHONE, 'required': False},
                {'prop_name': 'publicKey', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.CryptographicKey, 'fetch': True, 'required': False},
                {'prop_name': 'verification', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.VerificationObjectIssuer, 'fetch': False, 'required': False},
                {'task_type': ISSUER_PROPERTY_DEPENDENCIES, 'messageLevel': MESSAGE_LEVEL_WARNING}
            ]
        elif class_name == OBClasses.ExpectedRecipientProfile:
            # For ephemeral profiles representing recipients, there are few required properties.
            self.validators = [
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': False},
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': False, 'many': True,
                 'must_contain_one': ['Issuer', 'Profile'], 'default': OBClasses.Profile},
                {'prop_name': 'name', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'description', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'image', 'prop_type': ValueTypes.ID, 'required': False,
                 'expected_class': OBClasses.Image, 'fetch': False, 'allow_data_uri': True},
                {'prop_name': 'url', 'prop_type': ValueTypes.URL, 'required': False, 'many': True},
                {'prop_name': 'email', 'prop_type': ValueTypes.EMAIL, 'required': False, 'many': True},
                {'prop_name': 'telephone', 'prop_type': ValueTypes.TELEPHONE, 'required': False, 'many': True},
                {'prop_name': 'publicKey', 'prop_type': ValueTypes.ID, 'many': True,
                 'expected_class': OBClasses.CryptographicKey, 'fetch': False, 'required': False},
                {'prop_name': 'verification', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.VerificationObjectIssuer, 'fetch': False, 'required': False},
                {'task_type': VERIFY_RECIPIENT_IDENTIFIER, 'prerequisites': ASSERTION_VERIFICATION_DEPENDENCIES}
            ]
        elif class_name == OBClasses.AlignmentObject:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE,
                 'many': True, 'required': False, 'default': OBClasses.AlignmentObject},
                {'prop_name': 'targetName', 'prop_type': ValueTypes.TEXT, 'required': True},
                {'prop_name': 'targetUrl', 'prop_type': ValueTypes.URL, 'required': True},
                {'prop_name': 'description', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'targetFramework', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'targetCode', 'prop_type': ValueTypes.TEXT, 'required': False},
            ]
        elif class_name == OBClasses.Criteria:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE,
                 'many': True, 'required': False, 'default': OBClasses.Criteria},
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': False},
                {'prop_name': 'narrative', 'prop_type': ValueTypes.MARKDOWN_TEXT, 'required': False},
                {'task_type': CRITERIA_PROPERTY_DEPENDENCIES}
            ]
        elif class_name == OBClasses.IdentityObject:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': True, 'many': False,
                 'must_contain_one': ['id', 'email', 'url', 'telephone']},
                {'prop_name': 'identity', 'prop_type': ValueTypes.IDENTITY_HASH, 'required': True},
                {'prop_name': 'hashed', 'prop_type': ValueTypes.BOOLEAN, 'required': True},
                {'prop_name': 'salt', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'task_type': IDENTITY_OBJECT_PROPERTY_DEPENDENCIES}
            ]
        elif class_name == OBClasses.Evidence:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'many': True,
                 'required': False, 'default': 'Evidence'},
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': False},
                {'prop_name': 'narrative', 'prop_type': ValueTypes.MARKDOWN_TEXT, 'required': False},
                {'prop_name': 'name', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'description', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'genre', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'audience', 'prop_type': ValueTypes.TEXT, 'required': False},
            ]
        elif class_name == OBClasses.Image:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'many': True,
                 'required': False, 'default': 'schema:ImageObject'},
                {'prop_name': 'id', 'prop_type': ValueTypes.DATA_URI_OR_URL, 'required': True},
                {'prop_name': 'caption', 'prop_type': ValueTypes.TEXT, 'required': False},
                {'prop_name': 'author', 'prop_type': ValueTypes.IRI, 'required': False},
                {'task_type': IMAGE_VALIDATION, 'prop_name': 'id', 'allow_data_uri': True}
            ]
        elif class_name == OBClasses.VerificationObjectAssertion:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': True, 'many': False,
                 'must_contain_one': ['HostedBadge', 'SignedBadge']},
                {'prop_name': 'creator', 'prop_type': ValueTypes.ID,
                 'expected_class': OBClasses.CryptographicKey, 'fetch': True, 'required': False,
                 'prerequisites': ASSERTION_VERIFICATION_DEPENDENCIES},
            ]
        elif class_name == OBClasses.VerificationObjectIssuer:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': False, 'many': True,
                 'default': 'VerificationObject'},
                {'prop_name': 'verificationProperty', 'prop_type': ValueTypes.COMPACT_IRI, 'required': False},
                {'prop_name': 'startsWith', 'prop_type': ValueTypes.URL, 'required': False},
                {'prop_name': 'allowedOrigins', 'prop_type': ValueTypes.URL_AUTHORITY, 'required': False,
                 'many': True}
            ]
        elif class_name == OBClasses.RevocationList:
            self.validators = [
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': True, 'many': True,
                 'must_contain_one': OBClasses.RevocationList},
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': False},
                {'task_type': VALIDATE_REVOCATIONLIST_ENTRIES}
            ]
        elif class_name == OBClasses.Endorsement:
            self.validators = [
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': True},
                {'prop_name': 'type', 'prop_type': ValueTypes.RDF_TYPE, 'required': True,
                 'many': True, 'must_contain_one': [OBClasses.Endorsement]},
                {'prop_name': 'claim', 'prop_type': ValueTypes.ID, 'required': True, 'allow_remote_url': False,
                 'fetch': False, 'allow_data_uri': False, 'expected_class': [OBClasses.EndorsementClaim, class_name],
                 'full_validate': False},
                {'prop_name': 'issuedOn', 'prop_type': ValueTypes.DATETIME, 'required': True},
                {'prop_name': 'issuer', 'prop_type': ValueTypes.ID,  # 'prerequisites': ['ENDO_FLATTEN_ISS'],
                 'expected_class': OBClasses.Profile, 'fetch': True, 'required': True},
                {'prop_name': 'verification'},
                # {'task_type': FLATTEN_EMBEDDED_RESOURCE, 'prop_name': 'issuer', 'node_class': OBClasses.BadgeClass,
                #  'task_key': 'ENDO_FLATTEN_ISS'}
            ]
        elif class_name == OBClasses.EndorsementClaim:
            self.validators = [
                {'prop_name': 'id', 'prop_type': ValueTypes.IRI, 'required': True},
                {'prop_name': 'endorsementComment', 'prop_type': ValueTypes.MARKDOWN_TEXT, 'required': False}
            ]
            return  # Because this class appears paired with another, no need to repeat the generic props below.
        else:
            raise NotImplementedError("Chosen OBClass {} not implemented yet.".format(class_name))

        if class_name in OBClasses.PRIMARY_OBJECTS:
            self.validators += [
                {'prop_name': '@language', 'prop_type': ValueTypes.LANGUAGE, 'required': False},
                {'prop_name': 'version', 'prop_type': ValueTypes.TEXT_OR_NUMBER, 'required': False},
                {'prop_name': 'related', 'prop_type': ValueTypes.ID, 'required': False, 'allow_remote_url': True,
                 'fetch': False, 'allow_data_uri': False, 'expected_class': class_name, 'full_validate': False,
                 'many': True},
                {'prop_name': 'endorsement', 'prop_type': ValueTypes.ID, 'required': False, 'allow_remote_url': True,
                 'fetch': True, 'allow_data_uri': False, 'expected_class': OBClasses.Endorsement, 'many': True,
                 'prerequisites': ['PRI_FLATTEN_ENDRS']},
                {'task_type': FLATTEN_EMBEDDED_RESOURCE, 'prop_name': 'endorsement', 'node_class': OBClasses.Endorsement,
                 'task_key': 'PRI_FLATTEN_ENDRS'}
            ]


def _get_validation_actions(node_class, depth, node_id=None, node_path=None):
    validators = ClassValidators(node_class).validators
    actions = []
    for validator in validators:
        if validator.get('prop_type') == ValueTypes.RDF_TYPE:
            action = add_task(VALIDATE_RDF_TYPE_PROPERTY, **validator)
        if validator.get('prop_type') in ValueTypes.PRIMITIVES:
            action = add_task(VALIDATE_PROPERTY, **validator)
        elif validator.get('task_type') in CLASS_VALIDATION_TASKS:
            action = add_task(validator['task_type'], **validator)
        else:
            continue

        if node_id:
            action['node_id'] = node_id
        else:
            action['node_path'] = node_path

        action['depth'] = depth + 1
        actions.append(action)
    return actions


def detect_and_validate_node_class(state, task_meta, **options):
    node_id = task_meta.get('node_id')
    node_path = task_meta.get('node_path')
    depth = task_meta.get('depth')
    max_depth = options.get('max_validation_depth')

    if depth <= max_depth:
        if node_id:
            node = get_node_by_id(state, node_id)
        else:
            node = get_node_by_path(state, node_path)

        declared_node_type = node.get('type')
        node_class = None

        for ob_class in OBClasses.ALL_CLASSES:
            if declared_node_type == ob_class:
                node_class = OBClasses.default_for(ob_class)
                break

        actions = _get_validation_actions(node_class, depth, node_id, node_path)

        # Filter list for related nodes down to props that exist and 'id'
        if task_meta.get('full_validate', True) is False:
            actions = [a for a in actions
                       if a.get('prop_name') == 'id'
                       or a.get('prop_name') is not None and node.get(a['prop_name']) is not None]

        return task_result(
            True, "Declared type on node {} is {}".format(abv_node(node_id, node_path), declared_node_type),
            actions
        )
    else:
        return task_result(
            True, "Reached max validation depth", []
        )


def validate_expected_node_class(state, task_meta, **options):
    node_id = task_meta.get('node_id')
    node_path = task_meta.get('node_path')
    depth = task_meta.get('depth')
    max_depth = options.get('max_validation_depth')
    actions = []

    if depth <= max_depth:
        node_classes = [OBClasses.default_for(c) for c in list_of(task_meta['expected_class'])]

        for node_class in node_classes:
            actions += _get_validation_actions(node_class, depth, node_id, node_path)

        # Filter list for related nodes down to props that exist and 'id'
        if task_meta.get('full_validate', True) is False:
            if node_id:
                node = get_node_by_id(state, node_id)
            else:
                node = get_node_by_path(state, node_path)
            actions = [a for a in actions
                       if a.get('prop_name', 'UNKNOWN') in ['id', 'endorsementComment']
                       or a.get('prop_name') is not None and node.get(a['prop_name']) is not None
                       or a.get('prop_name') is None and a.get('run_non_core') is True]

        return task_result(
            True, "Queued property validations for class {} instance {}".format(
                node_class, abv_node(node_id, node_path)),
            actions
        )
    else:
        return task_result(
            True, "Reached max validation depth", actions
        )


"""
Class Validation Tasks
"""


def identity_object_property_dependencies(state, task_meta, **options):
    node_id = task_meta.get('node_id')
    node_path = task_meta.get('node_path')
    if node_id:
        node = get_node_by_id(state, node_id)
    else:
        node = get_node_by_path(state, node_path)

    node_class = task_meta.get('node_class')
    identity = node.get('identity')
    is_hashed = PrimitiveValueValidator.is_hashed_identity_hash(identity)
    is_email = bool(re.match(r'[^@]+@[^@]+$', identity))

    if node.get('hashed') and not is_hashed:
        return task_result(
            False,
            "Identity {} must match known hash style if hashed is true".format(identity))
    elif is_hashed and not node.get('hashed'):
        return task_result(
            False,
            "Identity {} must not be hashed if hashed is false".format(identity)
        )
    if not node.get('hashed') and 'email' in node.get('type') and not is_email:
        return task_result(False, "Email type identity must match email format.")

    return task_result(True, "IdentityObject passes validation rules.")


def criteria_property_dependencies(state, task_meta, **options):
    try:
        node_id = task_meta.get('node_id')
        node_path = task_meta.get('node_path')
        if node_id:
            node = get_node_by_id(state, node_id)
        else:
            node = get_node_by_path(state, node_path)
            node_id = node.get('id', '')
    except (IndexError, KeyError, TypeError):
        raise TaskPrerequisitesError()

    is_blank_id_node = bool(re.match(r'_:b\d+$', node_id))
    if is_blank_id_node and not node.get('narrative'):
        return task_result(False,
                           "Criteria node {} has no narrative. Either external id or narrative is required.".format(
                               abv_node(node_id, node_path))
                           )
    elif is_blank_id_node:
        return task_result(
            True, "Criteria node {} is a narrative-based piece of evidence.".format(
                abv_node(node_id, node_path)
            )
        )
    elif not is_blank_id_node and node.get('narrative'):
        return task_result(
            True, "Criteria node {} has a URL and narrative.".format(abv_node(node_id, node_path))
        )
    # Case to handle no narrative but other props preventing compaction down to simple id string:
    # {'id': 'http://example.com/1', 'name': 'Criteria Name'}
    return task_result(True, "Criteria node {} has a URL.".format(abv_node(node_id, node_path)))


def assertion_verification_dependencies(state, task_meta, **options):
    """
    Performs and/or queues some security checks for hosted assertions.
    """
    try:
        assertion_id = task_meta.get('node_id')
        node_path = task_meta.get('node_path')
        if assertion_id:
            node = get_node_by_id(state, assertion_id)
        elif node_path:
            node = get_node_by_path(state, node_path)
        node_id = node['verification']
        if isinstance(node_id, six.string_types):
            node = get_node_by_id(state, node_id)
        else:
            node = node_id
    except (IndexError, KeyError,):
        raise TaskPrerequisitesError()

    actions = []

    if node.get('type') == 'HostedBadge':
        actions.append(add_task(HOSTED_ID_IN_VERIFICATION_SCOPE, node_id=assertion_id))

    return task_result(
        True, '{} Assertion {} verification dependencies noted.'.format(
            node.get('type'), assertion_id),
        actions
    )


def assertion_timestamp_checks(state, task_meta, **options):
    try:
        node_id = task_meta['node_id']
        assertion = get_node_by_id(state, node_id)
        issued_on = aniso8601.parse_datetime(assertion['issuedOn'])
    except (IndexError, KeyError, ValueError,):
        raise TaskPrerequisitesError(task_meta)

    now = datetime.now(utc)
    if issued_on > now:
        return task_result(
            False, "Assertion {} has issue date {} in the future.".format(node_id, issued_on))

    if assertion.get('expires'):
        expires = aniso8601.parse_datetime(assertion['expires'])
        if expires < issued_on:
            return task_result(
                False, "Assertion {} expiration is prior to issue date.".format(node_id))

        if expires < now:
            return task_result(
                False, "Assertion {} expired on {}".format(node_id, assertion['expires'])
            )

    return task_result(
        True, "Assertion {} was issued and has not expired.".format(node_id))


def issuer_property_dependencies(state, task_meta, **options):
    try:
        node_id = task_meta['node_id']
        issuer_node = get_node_by_id(state, node_id)
    except (IndexError, KeyError,):
        raise TaskPrerequisitesError()

    if not bool(re.match('^http(s)?://', node_id)):
        return task_result(
            False, "Issuer Profile {} not hosted with HTTP-based identifier.".format(node_id) +
                   "Many platforms can only handle HTTP(s)-hosted issuers.")
    return task_result(True, "Issuer profile id meets expectations.")


def placeholder_task(state, task_meta, **options):
    return task_result(True, "Placeholder (prerequisite) task complete.")


def validate_revocationlist_entries(state, task_meta, **options):
    try:
        node_id = task_meta['node_id']
        revocation_list = get_node_by_id(state, node_id)
    except (IndexError, KeyError,):
        raise TaskPrerequisitesError()

    try:
        revoked_assertions = list_of(revocation_list['revokedAssertions'])
    except KeyError:
        return task_result(False, "RevocationList {} missing required property revokedAssertions.")

    for entry in revoked_assertions:
        if isinstance(entry, dict):
            try:
                if not PrimitiveValueValidator(ValueTypes.IRI)(entry['id']):
                    return task_result(False, "RevocationList {} has entry with id {} not in IRI format".format(
                        node_id, entry['id']
                    ))
            except KeyError:
                if not isinstance(entry.get('uid'), six.string_types):
                    return task_result(False, "RevocationList {} has entry with uid '{}' not in text format.".format(
                        node_id, abv(entry.get('uid'))
                    ))
        elif isinstance(entry, six.string_types):
            if not PrimitiveValueValidator(ValueTypes.IRI)(entry):
                return task_result(False, "RevocationList {} has entry with id {} not in IRI format".format(
                    node_id, entry
                ))
        else:
            return task_result(False, "RevocationList {} has entry with id {} not in IRI format".format(
                node_id, entry
            ))

    # Node flattening system will insert all entries into graph.
    return task_result(True)



import hashlib
import json
import responses
import unittest

from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V2_DICT
from openbadges.verifier.state import filter_failed_tasks
from openbadges.verifier.tasks.graph import jsonld_compact_data
from openbadges.verifier.tasks.task_types import JSONLD_COMPACT_DATA, VERIFY_RECIPIENT_IDENTIFIER
from openbadges.verifier.tasks.utils import filter_tasks
from openbadges.verifier.tasks.validation import OBClasses
from openbadges.verifier.tasks.verification import verify_recipient_against_trusted_profile
from openbadges.verifier.verifier import verification_store

try:
    from .testfiles.test_components import test_components
    from tests.utils import set_up_image_mock
except (ImportError, SystemError):
    from .testfiles.test_components import test_components
    from tests.utils import set_up_image_mock


class RecipientProfileVerificationTests(unittest.TestCase):
    def test_jsonld_compact_recipient_profile(self):
        recipient_profile = {'@context': OPENBADGES_CONTEXT_V2_DICT, 'email': 'nobody@example.com'}
        task_meta = add_task(
            JSONLD_COMPACT_DATA,
            data=json.dumps(recipient_profile),
            expected_class=OBClasses.ExpectedRecipientProfile)
        state = {}

        result, message, actions = jsonld_compact_data(state, task_meta)
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)

    @responses.activate
    def test_verify_badge_against_expected_profile(self):
        recipient_profile = {'email': 'nobody@example.org'}
        url = 'https://example.org/beths-robotics-badge.json'
        assertion = json.loads(test_components['2_0_basic_assertion'])
        assertion['recipient']['identity'] = 'sha256$' + hashlib.sha256(
            recipient_profile['email'].encode('utf8') + assertion['recipient']['salt'].encode('utf8')).hexdigest()

        responses.add(
            responses.GET, url, body=json.dumps(assertion), status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock('https://example.org/beths-robot-badge.png')
        responses.add(
            responses.GET, 'https://w3id.org/openbadges/v2',
            body=test_components['openbadges_context'], status=200,
            content_type='application/ld+json'
        )
        responses.add(
            responses.GET, 'https://example.org/robotics-badge.json',
            body=test_components['2_0_basic_badgeclass'], status=200,
            content_type='application/ld+json'
        )
        set_up_image_mock('https://example.org/robotics-badge.png')
        responses.add(
            responses.GET, 'https://example.org/organization.json',
            body=test_components['2_0_basic_issuer'], status=200,
            content_type='application/ld+json'
        )

        result = verification_store(url, recipient_profile)
        state = result.get_state()
        profile_verification_tasks = filter_tasks(state, name=VERIFY_RECIPIENT_IDENTIFIER)

        self.assertEqual(len(filter_failed_tasks(state)), 0)
        self.assertEqual(len(profile_verification_tasks), 1)
        self.assertTrue(profile_verification_tasks[0]['success'])

        recipient_profile = {'email': ['otheremail@example.org', 'nobody@example.org']}
        result = verification_store(url, recipient_profile)
        state = result.get_state()
        profile_verification_tasks = filter_tasks(state, name=VERIFY_RECIPIENT_IDENTIFIER)
        self.assertEqual(len(filter_failed_tasks(state)), 0)
        self.assertEqual(len(profile_verification_tasks), 1)
        self.assertTrue(profile_verification_tasks[0]['success'])

    def test_profile_with_salted_hashed_email(self):
        recipient_profile = {'id': '_:b0', 'email': 'nobody@example.org'}
        assertion = {
            'id': 'http://example.org',
            'type': 'Assertion',
            'recipient': '_:b1'
        }
        identity_object = {
            'id': '_:b1',
            'type': 'email',
            'hashed': True,
            'salt': 'Maldon',
            'identity': 'sha256$' + hashlib.sha256(recipient_profile['email'].encode('utf8') + 'Maldon'.encode('utf8')).hexdigest()
        }

        state = {'graph': [recipient_profile, assertion, identity_object]}
        task_meta = add_task(VERIFY_RECIPIENT_IDENTIFIER, node_id='_:b0')

        result, message, actions = verify_recipient_against_trusted_profile(state, task_meta)
        self.assertTrue(result)
        self.assertIn('nobody@example.org', message)

    def test_profile_with_salted_hashed_email_uppercase_hash(self):
        # Some implementations used an uppercase output of the sha256 hashing algorithm, which is equivalent.
        recipient_profile = {'id': '_:b0', 'email': 'nobody@example.org'}
        assertion = {
            'id': 'http://example.org',
            'type': 'Assertion',
            'recipient': '_:b1'
        }
        identity_object = {
            'id': '_:b1',
            'type': 'email',
            'hashed': True,
            'salt': 'Maldon',
            'identity': 'sha256$' + hashlib.sha256(recipient_profile['email'].encode('utf8') + 'Maldon'.encode('utf8')).hexdigest().upper()
        }

        state = {'graph': [recipient_profile, assertion, identity_object]}
        task_meta = add_task(VERIFY_RECIPIENT_IDENTIFIER, node_id='_:b0')

        result, message, actions = verify_recipient_against_trusted_profile(state, task_meta)
        self.assertTrue(result)
        self.assertIn('nobody@example.org', message)

    def test_unknown_identity_type(self):
        recipient_profile = {'id': '_:b0', 'schema:duns': '999999999'}
        assertion = {
            'id': 'http://example.org',
            'type': 'Assertion',
            'recipient': '_:b1'
        }
        identity_object = {
            'id': '_:b1',
            'type': 'schema:duns',
            'hashed': True,
            'salt': 'HimalayanPink',
            'identity': 'sha256$' + hashlib.sha256(
                recipient_profile['schema:duns'].encode('utf8') + 'HimalayanPink'.encode('utf8')).hexdigest()
        }

        state = {'graph': [recipient_profile, assertion, identity_object]}
        task_meta = add_task(VERIFY_RECIPIENT_IDENTIFIER, node_id='_:b0')

        result, message, actions = verify_recipient_against_trusted_profile(state, task_meta)
        self.assertTrue(result)
        self.assertIn(recipient_profile['schema:duns'], message)
        self.assertEqual(len(actions), 2)
        self.assertIn('schema:duns', actions[0]['message'], "Non-standard identifier reported")

    def test_profile_with_multiple_emails(self):
        recipient_profile = {'id': '_:b0', 'email': ['nobody@example.org', 'myaltemail@example.org']}
        assertion = {
            'id': 'http://example.org',
            'type': 'Assertion',
            'recipient': '_:b1'
        }
        identity_object = {
            'id': '_:b1',
            'type': 'email',
            'hashed': False,
            'identity': 'nobody@example.org'
        }

        state = {'graph': [recipient_profile, assertion, identity_object]}
        task_meta = add_task(VERIFY_RECIPIENT_IDENTIFIER, node_id='_:b0')

        result, message, actions = verify_recipient_against_trusted_profile(state, task_meta)
        self.assertTrue(result)
        self.assertIn('nobody@example.org', message)



test_components = {
'1_0_basic_assertion': """{
    "uid":"123abc",
    "recipient": {"identity": "test@example.com","hashed": false, "type": "email"},
    "badge": "http://a.com/badgeclass",
    "issuedOn": "2015-04-30",
    "verify": {"type": "hosted", "url": "http://a.com/instance"}
}""",
'1_0_basic_badgeclass': """{
    "name": "Basic Badge",
    "description": "Basic as it gets. v1.0",
    "image": "http://a.com/badgeclass_image",
    "criteria": "http://a.com/badgeclass_criteria",
    "issuer": "http://a.com/issuer"
}""",
'1_0_basic_issuer': """{
    "name": "Basic Issuer",
    "url": "http://a.com/issuer"
}""",
'1_0_assertion_with_errors': """{
    "uid":"123abc",
    "recipient": "test@example.com",
    "badge": "http://a.com/badgeclass",
    "issuedOn": "2015-04-30",
    "verify": {"type": "hosted", "url": "http://a.com/instance"}
}""",
'0_5_assertion': """{
    "recipient": "test@example.com",
    "badge": {
        "version": "0.5.0",
        "name": "Basic McBadge",
        "image": "http://oldstyle.com/images/1",
        "description": "A basic badge.",
        "criteria": "http://oldsyle.com/criteria/1",
        "issuer": {
            "origin": "http://oldstyle.com",
            "name": "Basic Issuer"
        }
    }
}""",
'0_5_1_assertion': """{
    "recipient": "sha256$85c4196c5516561cef673642157499b70066cb1070852b2a37fdbf3cc599b087",
    "salt": "sel gris",
    "issued_on": "2011-06-01",
    "badge": {
        "version": "0.5.0",
        "name": "Basic McBadge",
        "image": "http://oldstyle.com/images/2",
        "description": "A basic badge.",
        "criteria": "http://oldsyle.com/criteria/2",
        "issuer": {
            "origin": "http://oldstyle.com",
            "name": "Basic Issuer"
        }
    }
}""",
'1_0_basic_assertion_with_extra_properties': """{
    "uid":"123abc",
    "recipient": {"identity": "test@example.com","hashed": false, "type": "email"},
    "badge": "http://a.com/badgeclass",
    "issuedOn": "2015-04-30",
    "verify": {"type": "hosted", "url": "http://a.com/instance3"},
    "snood":"a very fun video game"
}""",
# Assertion awarded to nobody@example.org
'1_1_basic_assertion': """{
  "@context": "https://w3id.org/openbadges/v1",
  "type": "Assertion",
  "id": "https://example.org/beths-robotics-badge.json",
  "recipient": {
    "type": "email",
    "hashed": true,
    "salt": "deadsea",
    "identity": "sha256$ecf5409f3f4b91ab60cc5ef4c02aef7032354375e70cf4d8e43f6a1d29891942"
  },
  "image": "https://example.org/beths-robot-badge.png",
  "evidence": "https://example.org/beths-robot-work.html",
  "issuedOn": "2016-12-31T23:59:59Z",
  "badge": "https://example.org/robotics-badge.json",
  "verification": {
    "type": "hosted"
  }
}""",
'1_1_basic_badgeclass': """{
  "@context": "https://w3id.org/openbadges/v1",
  "type": "BadgeClass",
  "id": "https://example.org/robotics-badge.json",
  "name": "Awesome Robotics Badge",
  "description": "For doing awesome things with robots that people think is pretty great.",
  "image": "https://example.org/robotics-badge.png",
  "criteria": "https://example.org/robotics-badge.html",
  "issuer": "https://example.org/organization.json"
}""",
'1_1_basic_issuer': """{
  "@context": "https://w3id.org/openbadges/v1",
  "type": "Issuer",
  "id": "https://example.org/organization.json",
  "name": "An Example Badge Issuer",
  "url": "https://example.org",
  "email": "contact@example.org"
}""",
# Assertion awarded to nobody@example.org
'2_0_basic_assertion': """{
  "@context": "https://w3id.org/openbadges/v2",
  "type": "Assertion",
  "id": "https://example.org/beths-robotics-badge.json",
  "recipient": {
    "type": "email",
    "hashed": true,
    "salt": "deadsea",
    "identity": "sha256$ecf5409f3f4b91ab60cc5ef4c02aef7032354375e70cf4d8e43f6a1d29891942"
  },
  "image": "https://example.org/beths-robot-badge.png",
  "evidence": "https://example.org/beths-robot-work.html",
  "issuedOn": "2016-12-31T23:59:59Z",
  "badge": "https://example.org/robotics-badge.json",
  "verification": {
    "type": "hosted"
  }
}""",
'2_0_basic_badgeclass': """{
  "@context": "https://w3id.org/openbadges/v2",
  "type": "BadgeClass",
  "id": "https://example.org/robotics-badge.json",
  "name": "Awesome Robotics Badge",
  "description": "For doing awesome things with robots that people think is pretty great.",
  "image": "https://example.org/robotics-badge.png",
  "criteria": "https://example.org/robotics-badge.html",
  "issuer": "https://example.org/organization.json"
}""",
'2_0_basic_issuer': """{
  "@context": "https://w3id.org/openbadges/v2",
  "type": "Issuer",
  "id": "https://example.org/organization.json",
  "name": "An Example Badge Issuer",
  "url": "https://example.org",
  "email": "contact@example.org"
}""",
'openbadges_context': """
{"@context": {"issuedOn": {"@id": "obi:issueDate", "@type": "xsd:dateTime"}, "AlignmentObject": "schema:AlignmentObject", "uid": {"@id": "obi:uid"}, "claim": {"@id": "cred:claim", "@type": "@id"}, "targetCode": {"@id": "obi:targetCode"}, "image": {"@id": "schema:image", "@type": "@id"}, "Endorsement": "cred:Credential", "Assertion": "obi:Assertion", "related": {"@id": "dc:relation", "@type": "@id"}, "evidence": {"@id": "obi:evidence", "@type": "@id"}, "sec": "https://w3id.org/security#", "Criteria": "obi:Criteria", "owner": {"@id": "sec:owner", "@type": "@id"}, "revocationList": {"@id": "obi:revocationList", "@type": "@id"}, "targetName": {"@id": "schema:targetName"}, "id": "@id", "alignment": {"@id": "obi:alignment", "@type": "@id"}, "allowedOrigins": {"@id": "obi:allowedOrigins"}, "Profile": "obi:Profile", "startsWith": {"@id": "http://purl.org/dqm-vocabulary/v1/dqm#startsWith"}, "author": {"@id": "schema:author", "@type": "@id"}, "FrameValidation": "obi:FrameValidation", "validationFrame": "obi:validationFrame", "creator": {"@id": "dc:creator", "@type": "@id"}, "validationSchema": "obi:validationSchema", "validatesType": "obi:validatesType", "version": {"@id": "schema:version"}, "BadgeClass": "obi:BadgeClass", "endorsement": {"@id": "cred:credential", "@type": "@id"}, "revocationReason": {"@id": "obi:revocationReason"}, "RevocationList": "obi:RevocationList", "issuer": {"@id": "obi:issuer", "@type": "@id"}, "type": "@type", "email": {"@id": "schema:email"}, "targetDescription": {"@id": "schema:targetDescription"}, "schema": "http://schema.org/", "targetUrl": {"@id": "schema:targetUrl"}, "criteria": {"@id": "obi:criteria", "@type": "@id"}, "verificationProperty": {"@id": "obi:verificationProperty"}, "description": {"@id": "schema:description"}, "Extension": "obi:Extension", "tags": {"@id": "schema:keywords"}, "CryptographicKey": "sec:Key", "expires": {"@id": "sec:expiration", "@type": "xsd:dateTime"}, "hosted": "obi:HostedBadge", "dc": "http://purl.org/dc/terms/", "telephone": {"@id": "schema:telephone"}, "publicKey": {"@id": "sec:publicKey", "@type": "@id"}, "badge": {"@id": "obi:badge", "@type": "@id"}, "endorsementComment": {"@id": "obi:endorsementComment"}, "genre": {"@id": "schema:genre"}, "hashed": {"@id": "obi:hashed", "@type": "xsd:boolean"}, "recipient": {"@id": "obi:recipient", "@type": "@id"}, "HostedBadge": "obi:HostedBadge", "identity": {"@id": "obi:identityHash"}, "revoked": {"@id": "obi:revoked", "@type": "xsd:boolean"}, "verify": "verification", "VerificationObject": "obi:VerificationObject", "name": {"@id": "schema:name"}, "publicKeyPem": {"@id": "sec:publicKeyPem"}, "obi": "https://w3id.org/openbadges#", "url": {"@id": "schema:url", "@type": "@id"}, "cred": "https://w3id.org/credentials#", "Image": "obi:Image", "created": {"@id": "dc:created", "@type": "xsd:dateTime"}, "IdentityObject": "obi:IdentityObject", "signed": "obi:SignedBadge", "Evidence": "obi:Evidence", "narrative": {"@id": "obi:narrative"}, "caption": {"@id": "schema:caption"}, "audience": {"@id": "obi:audience"}, "extensions": "https://w3id.org/openbadges/extensions#", "verification": {"@id": "obi:verify", "@type": "@id"}, "xsd": "http://www.w3.org/2001/XMLSchema#", "TypeValidation": "obi:TypeValidation", "revokedAssertions": {"@id": "obi:revoked"}, "SignedBadge": "obi:SignedBadge", "validation": "obi:validation", "salt": {"@id": "obi:salt"}, "targetFramework": {"@id": "schema:targetFramework"}, "Issuer": "obi:Issuer"}}
""",
'openbadges_context_v1': """{
  "@context": [
  {
    "id": "@id",
    "type": "@type",

    "ob": "https://w3id.org/openbadges#",
    "dc": "http://purl.org/dc/terms/",
    "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
    "rdfs": "http://www.w3.org/2000/01/rdf-schema#",
    "sec": "https://w3id.org/security#",
    "schema": "http://schema.org/",
    "xsd": "http://www.w3.org/2001/XMLSchema#",

    "about": {"@id": "schema:about", "@type": "@id"},
    "alignment": {"@id": "ob:alignment", "@type": "@id"},
    "badge": {"@id": "ob:badge", "@type": "@id"},
    "badgeOffer": {"@id": "ob:badgeOffer", "@type": "@id"},
    "badgeTemplate": {"@id": "ob:badgeTemplate", "@type": "@id"},
    "criteria": {"@id": "ob:criteria", "@type": "@id"},
    "evidence": {"@id": "ob:evidence", "@type": "@id"},
    "issued": {"@id": "ob:issued", "@type": "xsd:dateTime"},
    "issuer": {"@id": "ob:issuer", "@type": "@id"},
    "recipient": {"@id": "ob:recipient", "@type": "@id"},
    "recipientEmail": "ob:recipientEmail",
    "recipientPassword": "ob:recipientPassword",
    "tag": "ob:tag",
    "Identity": "ob:Identity",
    "Badge": "ob:Badge",
    "BadgeOffer": "ob:BadgeOffer",
    "BadgeTemplate": "ob:BadgeTemplate",

    "address": {"@id": "schema:address", "@type": "@id"},
    "addressCountry": "schema:addressCountry",
    "addressLocality": "schema:addressLocality",
    "addressRegion": "schema:addressRegion",
    "comment": "rdfs:comment",
    "created": {"@id": "dc:created", "@type": "xsd:dateTime"},
    "creator": {"@id": "dc:creator", "@type": "@id"},
    "description": "schema:description",
    "email": "schema:email",
    "familyName": "schema:familyName",
    "givenName": "schema:givenName",
    "image": {"@id": "schema:image", "@type": "@id"},
    "label": "rdfs:label",
    "name": "schema:name",
    "postalCode": "schema:postalCode",
    "streetAddress": "schema:streetAddress",
    "title": "dc:title",
    "url": {"@id": "schema:url", "@type": "@id"},
    "PostalAddress": "schema:PostalAddress",

    "identityService": {"@id": "https://w3id.org/identity#identityService", "@type": "@id"},

    "credential": {"@id": "sec:credential", "@type": "@id"},
    "cipherAlgorithm": "sec:cipherAlgorithm",
    "cipherData": "sec:cipherData",
    "cipherKey": "sec:cipherKey",
    "claim": {"@id": "sec:claim", "@type": "@id"},
    "digestAlgorithm": "sec:digestAlgorithm",
    "digestValue": "sec:digestValue",
    "domain": "sec:domain",
    "expires": {"@id": "sec:expiration", "@type": "xsd:dateTime"},
    "initializationVector": "sec:initializationVector",
    "nonce": "sec:nonce",
    "normalizationAlgorithm": "sec:normalizationAlgorithm",
    "owner": {"@id": "sec:owner", "@type": "@id"},
    "password": "sec:password",
    "privateKey": {"@id": "sec:privateKey", "@type": "@id"},
    "privateKeyPem": "sec:privateKeyPem",
    "publicKey": {"@id": "sec:publicKey", "@type": "@id"},
    "publicKeyPem": "sec:publicKeyPem",
    "publicKeyService": {"@id": "sec:publicKeyService", "@type": "@id"},
    "revoked": {"@id": "sec:revoked", "@type": "xsd:dateTime"},
    "signature": "sec:signature",
    "signatureAlgorithm": "sec:signatureAlgorithm",
    "signatureValue": "sec:signatureValue",
    "EncryptedMessage": "sec:EncryptedMessage",
    "CryptographicKey": "sec:Key",
    "GraphSignature2012": "sec:GraphSignature2012"
  },
  {
    "id": "@id",
    "type": "@type",

    "obi": "https://w3id.org/openbadges#",
    "extensions": "https://w3id.org/openbadges/extensions#",
    "validation": "obi:validation",

    "xsd": "http://www.w3.org/2001/XMLSchema#",
    "schema": "http://schema.org/",
    "sec": "https://w3id.org/security#",

    "Assertion": "obi:Assertion",
    "BadgeClass": "obi:BadgeClass",
    "Issuer": "obi:Issuer",
    "IssuerOrg": "obi:Issuer",
    "Extension": "obi:Extension",
    "hosted": "obi:HostedBadge",
    "signed": "obi:SignedBadge",
    "TypeValidation": "obi:TypeValidation",
    "FrameValidation": "obi:FrameValidation",

    "name": { "@id": "schema:name" },
    "description": { "@id": "schema:description" },
    "url": { "@id": "schema:url", "@type": "@id" },
    "image": { "@id": "schema:image", "@type": "@id" },
 
    "uid": { "@id": "obi:uid" },
    "recipient": { "@id": "obi:recipient", "@type": "@id" },
    "hashed": { "@id": "obi:hashed", "@type": "xsd:boolean" },
    "salt": { "@id": "obi:salt" },
    "identity": { "@id": "obi:identityHash" },
    "issuedOn": { "@id": "obi:issueDate", "@type": "xsd:dateTime" },
    "expires": { "@id": "sec:expiration", "@type": "xsd:dateTime" },
    "evidence": { "@id": "obi:evidence", "@type": "@id" },
    "verify": { "@id": "obi:verify", "@type": "@id" },

    "badge": { "@id": "obi:badge", "@type": "@id" },
    "criteria": { "@id": "obi:criteria", "@type": "@id" },
    "tags": { "@id": "schema:keywords" },
    "alignment": { "@id": "obi:alignment", "@type": "@id" },

    "issuer": { "@id": "obi:issuer", "@type": "@id" },
    "email": "schema:email",
    "revocationList": { "@id": "obi:revocationList", "@type": "@id" },

    "validatesType": "obi:validatesType",
    "validationSchema": "obi:validationSchema",
    "validationFrame": "obi:validationFrame"
  }],

"validation": [
  {
    "type": "TypeValidation",
    "validatesType": "Assertion",
    "validationSchema": "https://openbadgespec.org/v1/schema/assertion.json"
  },
    {
      "type": "TypeValidation",
      "validatesType": "BadgeClass",
      "validationSchema": "https://openbadgespec.org/v1/schema/badgeclass.json"
    },
    {
      "type": "TypeValidation",
      "validatesType": "Issuer",
      "validationSchema": "https://openbadgespec.org/v1/schema/issuer.json"
    },
    {
      "type": "TypeValidation",
      "validatesType": "Extension",
      "validationSchema": "https://openbadgespec.org/v1/schema/extension.json"
    }
  ]
}
"""
}



from ..actions.action_types import STORE_INPUT, SET_INPUT_TYPE, STORE_ORIGINAL_RESOURCE


def input_reducer(state=None, action=None):
    if isinstance(state, dict):
        new_state = state.copy()
    else:
        new_state = {}

    if action.get('type') == STORE_INPUT:
        new_state.update({'value': action.get('input')})
    elif action.get('type') == SET_INPUT_TYPE:
        new_state.update({'input_type': action.get('input_type')})
    elif action.get('type') == STORE_ORIGINAL_RESOURCE:
        original_json = new_state.get('original_json', {})
        original_json[action['node_id']] = action.get('data')
        new_state['original_json'] = original_json

    return new_state



import os
from setuptools import find_packages, setup


# Build README info
short_description = 'A python module that performs verification for Open Badges.'
try:
    import pypandoc
    pypandoc.convert_file('README.md', 'rst', outputfile='README.rst')
    with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
        README = readme.read()
except (ImportError, RuntimeError, OSError):
    README = short_description

# import VERSION
try:
    execfile(os.path.join(os.path.dirname(__file__), 'openbadges/version.py'))
except NameError:
    exec(open(os.path.join(os.path.dirname(__file__), 'openbadges/version.py')).read())

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name='badgecheck',
    version=".".join(map(str, VERSION)),
    packages=find_packages(exclude=['tests', 'tests.*']),
    include_package_data=True,
    license='Apache 2',
    description=short_description,
    long_description=README,
    url='https://github.com/IMSGlobal/openbadges-validator-core',
    author='IMS Global',
    author_email='openbadgesinfo@imsglobal.org',
    classifiers=[
        'Environment :: Console',
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Education',
        'Topic :: Utilities',
        'Intended Audience :: Developers'
    ],
    install_requires=[
        'aniso8601>=1.2.0',
        'Click == 6.7',
        'future>=0.16.0',
        'jsonschema==2.6.0',
        'language-tags==0.4.3',
        'openbadges-bakery>=1.1.0',
        'pycryptodome==3.6.6',
        'pydux==0.2.2',
        'PyLD==0.7.1',
        'python-jose==3.0.1',
        'python-mimeparse==1.6.0',
        'pytz==2017.2',
        'requests >= 2.13',
        'requests_cache>=0.4.13',
        'rfc3986==0.4.1',
        'validators==0.11.2',
    ],
    extras_require={
        'server':  ["Flask==0.12.1", 'gunicorn==19.7.1'],
    },
    entry_points="""
        [console_scripts]
        openbadges=openbadges.command_line:cli
    """
)



import json
import jsonschema
from pyld import jsonld

from ..actions.tasks import add_task
from ..exceptions import TaskPrerequisitesError
from ..openbadges_context import OPENBADGES_CONTEXT_V2_URI
from ..state import get_node_by_id, get_node_by_path
from ..utils import jsonld_use_cache, list_of

from .task_types import VALIDATE_EXTENSION_SINGLE
from .utils import (abbreviate_value as abv,
                    abbreviate_node_id as abv_node,
                    combine_contexts, is_ld_term_in_list, task_result,)


def validate_single_extension(state, task_meta, **options):
    # node, extension, node_json=None, node_id_string=None, context_urls=None
    try:
        extension = task_meta['extension']

        node_id = task_meta.get('node_id')
        node_path = task_meta.get('node_path')
        if node_id:
            node = get_node_by_id(state, node_id)
        else:
            node = get_node_by_path(state, node_path)
        if not node:
            node = json.loads(task_meta['node_json'])

        node_id_string = abv_node(node_id, node_path)
        if node_id_string is None:
            node_id_string = node.get('id', 'unknown node')
    except (IndexError, TypeError, KeyError):
        raise TaskPrerequisitesError()

    node_data = node.copy()

    # Validate against JSON-schema
    context = extension['context_json']
    extension_type = extension['validates_type']
    schema = extension['validation_schema']

    node_data['@context'] = OPENBADGES_CONTEXT_V2_URI
    compact_data = jsonld.compact(
        node_data, {'@context': combine_contexts(OPENBADGES_CONTEXT_V2_URI, context)},
        options=options.get('jsonld_options', jsonld_use_cache))

    try:
        jsonschema.validate(compact_data, schema)
    except jsonschema.ValidationError as e:
        return task_result(
            False, "Extension {} did not validate on node {}: {}".format(
                extension_type, node_id_string, e.message
            )
        )

    return task_result(True, "Extension {} validated on node {}".format(
        extension_type, node_id_string
    ))


def validate_extension_node(state, task_meta, **options):
    try:
        node_id = task_meta.get('node_id')
        node_path = task_meta.get('node_path')
        context_urls = task_meta.get('context_urls')
        node_types = list_of(task_meta.get('types_to_test', []))
        if node_id:
            node = get_node_by_id(state, node_id)
        else:
            node = get_node_by_path(state, node_path)

        node_json = task_meta.get('node_json')  # Ok to be None
    except (KeyError, ValueError, IndexError, TypeError):
        raise TaskPrerequisitesError()

    if not context_urls:
        return task_result(False, "Could not determine extension type to test: no contexts defined")

    if not node_types:
        node_types = [t for t in node['type'] if t != 'Extension']

    jsonld_options = options.get('jsonld_options', jsonld_use_cache)
    loader = jsonld_options['documentLoader']

    extensions_to_test = []
    for context_url in context_urls:
        if context_url == OPENBADGES_CONTEXT_V2_URI:
            continue

        response = loader.session.get(context_url, headers={'Accept': 'application/ld+json, application/json'})
        try:
            context_json = response.json()
        except TypeError:
            continue
        context_compact = jsonld.compact(context_json, OPENBADGES_CONTEXT_V2_URI, options=jsonld_options)

        validation = list_of(context_compact.get('validation'))
        for val_entry in validation:
            if is_ld_term_in_list(
                    val_entry.get('validatesType'), node_types,
                    options=options.get('jsonld_options', jsonld_use_cache)):
                try:
                    schema_url = val_entry['validationSchema']
                    schema_json = loader.session.get(
                        schema_url, headers={'Accept': 'application/ld+json, application/json'}).json()
                except TypeError:
                    return task_result(False, 'Could not load JSON-schema from URL {}'.format(abv(schema_url)))

                extensions_to_test.append({
                    'context_url': context_url,
                    'context_json': context_json,
                    'validates_type': val_entry['validatesType'],
                    'validation_schema': schema_json
                })

    if not extensions_to_test:
        return task_result(False, "Could not determine extension type to test")
    elif len(extensions_to_test) > 1:
        # If there is more than one extension, return each validation as a separate task
        actions = [
            add_task(VALIDATE_EXTENSION_SINGLE, node_id=node_id, node_path=node_path,
                     node_json=node_json, extension=t)
            for t in extensions_to_test
        ]
        return task_result(
            True, "Multiple extension types {} discovered in node {}".format(
                abv([e['validates_type'] for e in extensions_to_test]), abv_node(node_id, node_path)
            ), actions)
    else:
        return validate_single_extension(
            state, add_task(
                VALIDATE_EXTENSION_SINGLE,
                node_id=node_id, node_path=node_path, node_json=node_json,
                extension=extensions_to_test[0], **options))



# -*- coding: utf-8 -*-
import json
import os
import responses
import unittest

from openbadges.verifier.actions.action_types import SET_OPENBADGES_VERSION, SET_VALIDATION_SUBJECT, REPORT_MESSAGE
from openbadges.verifier.actions.tasks import add_task
from openbadges.verifier.actions.input import store_input
from openbadges.verifier.openbadges_context import OPENBADGES_CONTEXT_V1_URI, OPENBADGES_CONTEXT_V2_URI
from openbadges.verifier.reducers import main_reducer
from openbadges.verifier.tasks.task_types import (DETECT_INPUT_TYPE, FETCH_HTTP_NODE, INTAKE_JSON, JSONLD_COMPACT_DATA,
                                                  UPGRADE_0_5_NODE, UPGRADE_1_0_NODE, UPGRADE_1_1_NODE)
from openbadges.verifier.tasks import run_task, task_named
from openbadges.verifier.state import INITIAL_STATE
from openbadges.verifier.tasks.validation import OBClasses
from openbadges.verifier.verifier import generate_report, verification_store, verify

from .testfiles.test_components import test_components


def setUpContextCache():
    v2_data = test_components['openbadges_context']
    responses.add(
        responses.GET, OPENBADGES_CONTEXT_V2_URI,
        body=v2_data, status=200, content_type='application/ld+json'
    )
    v1_data = test_components['openbadges_context_v1']
    responses.add(
        responses.GET, OPENBADGES_CONTEXT_V1_URI,
        body=v1_data, status=200, content_type='application/ld+json'
    )


class V1_1DetectionAndUpgradesTests(unittest.TestCase):
    def test_can_detect_v1_based_on_context(self):
        json_data = test_components['1_1_basic_assertion']
        state = INITIAL_STATE

        task = add_task(INTAKE_JSON, data=json_data)
        result, message, actions = task_named(task['name'])(state, task)
        self.assertTrue(result)
        self.assertEqual(actions[2]['name'], UPGRADE_1_1_NODE)

        json_data = test_components['1_1_basic_badgeclass']
        task['data'] = json_data
        result, message, actions = task_named(task['name'])(state, task)
        self.assertTrue(result)
        self.assertEqual(actions[2]['name'], UPGRADE_1_1_NODE)

        json_data = test_components['1_1_basic_issuer']
        task['data'] = json_data
        result, message, actions = task_named(task['name'])(state, task)
        self.assertTrue(result)
        self.assertEqual(actions[2]['name'], UPGRADE_1_1_NODE)

        json_data = test_components['1_0_basic_issuer']
        task['data'] = json_data
        result, message, actions = task_named(task['name'])(state, task)
        self.assertTrue(result)
        self.assertEqual(actions[1]['name'], UPGRADE_1_0_NODE)

    @responses.activate
    def test_upgrade_1_1_assertion(self):
        setUpContextCache()
        json_data = test_components['1_1_basic_assertion']
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, node_id='https://example.org/beths-robotics-badge.json', data=json_data)

        result, message, actions = task_named(INTAKE_JSON)(state, task)
        for action in actions:
            state = main_reducer(state, action)
        for task in state.get('tasks'):
            result, message, actions = task_named(task['name'])(state, task)
            for action in actions:
                state = main_reducer(state, action)

        result, message, actions = task_named(UPGRADE_1_1_NODE)(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 0)

        # Test timestamp upgrading
        state['graph'][0]['issuedOn'] = 1500423730
        result, message, actions = task_named(UPGRADE_1_1_NODE)(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 1)
        self.assertEqual(actions[0]['data']['issuedOn'], '2017-07-19T00:22:10+00:00')
        self.assertEqual(len(actions[0]['data'].keys()), 1, "There is a patch made of one prop")

        state['graph'][0]['issuedOn'] = '1500423730.5'
        result, message, actions = task_named(UPGRADE_1_1_NODE)(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 1)
        state = main_reducer(state, actions[0])

        state['graph'][0]['issuedOn'] = '2016-05-15'
        result, message, actions = task_named(UPGRADE_1_1_NODE)(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 1)
        state = main_reducer(state, actions[0])

    @responses.activate
    def test_upgrade_1_1_badgeclass(self):
        setUpContextCache()
        json_data = test_components['1_1_basic_badgeclass']
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, node_id='https://example.org/robotics-badge.json', data=json_data)

        result, message, actions = task_named(INTAKE_JSON)(state, task)
        for action in actions:
            state = main_reducer(state, action)
        for task in state.get('tasks'):
            result, message, actions = task_named(task['name'])(state, task)
            for action in actions:
                state = main_reducer(state, action)

        # Test criteria class upgrade
        state['graph'][0]['alignment'] = {
            'url': 'http://somewhere.overtherainbow.net/wayuphigh',
            'name': "Knowledge of children's songs"
        }
        result, message, actions = task_named(UPGRADE_1_1_NODE)(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 1)

    @responses.activate
    def test_upgrade_1_1_badgeclass_with_multiple_alignments(self):
        setUpContextCache()
        data = json.loads(test_components['1_1_basic_badgeclass'])
        data['alignment'] = [
            {
                'name': 'Target 1 Name',
                'description': 'Target 1 Description',
                'url': 'http://example.org/target-1'
            },
            {
                'name': 'Target 2 Name',
                'url': 'http://example.org/target-2'
            }
        ]
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, node_id='https://example.org/robotics-badge.json', data=json.dumps(data))

        result, message, actions = task_named(INTAKE_JSON)(state, task)
        for action in actions:
            state = main_reducer(state, action)
        for task in state.get('tasks'):
            result, message, actions = task_named(task['name'])(state, task)
            for action in actions:
                state = main_reducer(state, action)

        # Test alignment object upgrades
        self.assertEqual(len(state['graph'][0]['alignment']), 2)
        self.assertEqual(state['graph'][0]['alignment'][0]['targetUrl'], data['alignment'][0]['url'])

    def test_upgrade_1_1_issuer(self):
        setUpContextCache()
        data = json.loads(test_components['1_1_basic_issuer'])
        data['type'] = 'IssuerOrg'  # Test alias that was accepted in v1.1 context
        json_data = json.dumps(data)
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, node_id='https://example.org/organization.json', data=json_data)

        result, message, actions = task_named(INTAKE_JSON)(state, task)
        for action in actions:
            state = main_reducer(state, action)
        for task in state.get('tasks'):
            result, message, actions = task_named(task['name'])(state, task)
            for action in actions:
                state = main_reducer(state, action)

        self.assertTrue(result)
        self.assertEqual(state['graph'][0]['type'], OBClasses.Issuer)

    @responses.activate
    def test_upgrade_1_1_issuer_in_full_verify_with_redirect(self):
        setUpContextCache()
        old_url = "http://example.org/oldissuerurl"
        data = {
            "@context": "https://w3id.org/openbadges/v1",
            "type": "Issuer",
            "id": "http://example.org/realissuerurl",
            "description": "Example Badge Issuer",
            "url": "http://example.org",
            "email": "noreply@example.org",
            "name": "Test Issuer",
            "image": "http://example.org/image"
        }

        responses.add(responses.GET, old_url, json=data)
        responses.add(responses.GET, data['id'], json=data)

        result = verify(old_url)
        self.assertTrue(result['report']['valid'])


class V1_0DetectionAndUpgradesTests(unittest.TestCase):
    def test_json_1_0_assertion_input(self):
        assertion_data = test_components['1_0_basic_assertion']
        state = INITIAL_STATE
        action = store_input(assertion_data)
        state = main_reducer(state, action)
        self.assertEqual(state['input']['value'], assertion_data)

        task_meta = add_task(DETECT_INPUT_TYPE)
        result, message, actions = run_task(state, task_meta)
        self.assertTrue(result)
        self.assertIn(SET_VALIDATION_SUBJECT, [a.get('type') for a in actions])
        self.assertIn(FETCH_HTTP_NODE, [a.get('name') for a in actions])

    def test_upgrade_1_0_assertion(self):
        json_data = test_components['1_0_basic_assertion']
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, node_id='http://a.com/instance', data=json_data)

        result, message, actions = task_named(INTAKE_JSON)(state, task)
        for action in actions:
            state = main_reducer(state, action)

        task = state.get('tasks')[0]
        result, message, actions = task_named(task['name'])(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)
        for action in actions:
            state = main_reducer(state, action)

        self.assertEqual(len(state.get('tasks')), 3)
        modified_data = json.loads(state.get('tasks')[1]['data'])
        self.assertEqual(modified_data['@context'], OPENBADGES_CONTEXT_V1_URI)
        self.assertEqual(modified_data['id'], modified_data['verify']['url'])
        self.assertEqual(modified_data['type'], OBClasses.Assertion)

    def test_upgrade_1_0_badgeclass(self):
        json_data = test_components['1_0_basic_badgeclass']
        node_id = 'http://a.com/badgeclass'
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, node_id=node_id, data=json_data)

        result, message, actions = task_named(INTAKE_JSON)(state, task)
        for action in actions:
            state = main_reducer(state, action)

        task = state.get('tasks')[0]
        result, message, actions = task_named(task['name'])(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)
        for action in actions:
            state = main_reducer(state, action)

        self.assertEqual(len(state.get('tasks')), 3)
        modified_data = json.loads(state.get('tasks')[1]['data'])
        self.assertEqual(modified_data['@context'], OPENBADGES_CONTEXT_V1_URI)
        self.assertEqual(modified_data['id'], node_id)
        self.assertEqual(modified_data['type'], OBClasses.BadgeClass)

    def test_upgrade_1_0_issuer(self):
        json_data = test_components['1_0_basic_issuer']
        node_id = 'http://a.com/issuer'
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, node_id=node_id, data=json_data)

        result, message, actions = task_named(INTAKE_JSON)(state, task)
        for action in actions:
            state = main_reducer(state, action)

        task = state.get('tasks')[0]
        result, message, actions = task_named(task['name'])(state, task)
        self.assertTrue(result)
        self.assertEqual(len(actions), 2)
        for action in actions:
            state = main_reducer(state, action)

        self.assertEqual(len(state.get('tasks')), 3)
        modified_data = json.loads(state.get('tasks')[1]['data'])
        self.assertEqual(modified_data['@context'], OPENBADGES_CONTEXT_V1_URI)
        self.assertEqual(modified_data['id'], node_id)
        self.assertEqual(modified_data['type'], OBClasses.Issuer)

    @responses.activate
    def full_validate_1_0_to_2_0_conversion(self):
        setUpContextCache()
        assertion_data = json.loads(test_components['1_0_basic_assertion_with_extra_properties'])
        badgeclass_data = json.loads(test_components['1_0_basic_badgeclass'])
        issuer_data = json.loads(test_components['1_0_basic_issuer'])

        responses.add(
            responses.GET, assertion_data['verify']['url'],
            json=assertion_data
        )
        responses.add(
            responses.GET, assertion_data['badge'],
            json=badgeclass_data
        )
        responses.add(
            responses.GET, badgeclass_data['issuer'],
            json=issuer_data
        )
        png_badge = os.path.join(os.path.dirname(__file__), 'testfiles', 'public_domain_heart.png')
        with open(png_badge, 'rb') as image:
            responses.add(
                responses.GET, badgeclass_data['image'], body=image.read(), status=200, content_type='image/png'
            )

        store = verification_store(assertion_data['verify']['url'])
        state = store.get_state()
        report = generate_report(store)

        self.assertTrue(report['report']['valid'])
        assertion_node = state['graph'][0]
        badgeclass_node = state['graph'][1]
        issuer_node = state['graph'][2]

        self.assertEqual(assertion_node['id'], assertion_data['verify']['url'])
        self.assertEqual(badgeclass_node['@context'], OPENBADGES_CONTEXT_V2_URI)
        self.assertEqual(issuer_node['type'], OBClasses.Issuer)

        self.assertEqual(report['report']['openBadgesVersion'], '1.0')


class V1_0DetectionAndUpgradeTests(unittest.TestCase):
    def setUp(self):
        self.assertion_data = {
            "uid": u"abc123™",
            "recipient": "sha256$a4a934a0bfc882a34a3e71650e40789453b2db9799a51a2d084a64caadd72397",
            "salt": "2e2bad0df9e11272ffbcee86e4c7edd4",
            "issued_on": "2017-01-01",
            "badge": {
                "name": "Test Badge for admin@bluemantis.com",
                "image": "http://example.org/image",
                "description": "Awarded using the php example codebase",
                "criteria": "http://example.org/criteria",
                "issuer": {
                    "origin": "example.org",
                    "name": u"Test Issuer™",
                    "org": None,
                    "contact": None
                }
            }
        }

    def test_can_detect_and_upgrade_v_0_5(self):
        assertion_data = self.assertion_data
        state = INITIAL_STATE
        task = add_task(INTAKE_JSON, data=json.dumps(assertion_data), node_id='http://example.org/assertion')

        result, message, actions = run_task(state, task)
        self.assertTrue(result)
        version_action = [a for a in actions if a.get('type') == SET_OPENBADGES_VERSION][0]
        self.assertEqual(version_action['version'], '0.5')
        upgrade_action = [a for a in actions if a.get('name') == UPGRADE_0_5_NODE][0]

        result, message, actions = run_task(state, upgrade_action)
        self.assertTrue(result)

        self.assertEqual(len(actions), 2)
        self.assertEqual(actions[0]['name'], JSONLD_COMPACT_DATA)

    def test_json_0_5_assertion_input(self):
        assertion_data = json.dumps(self.assertion_data)
        state = INITIAL_STATE
        action = store_input(assertion_data)
        state = main_reducer(state, action)
        self.assertEqual(state['input']['value'], assertion_data)

        task_meta = add_task(DETECT_INPUT_TYPE)
        result, message, actions = run_task(state, task_meta)
        self.assertTrue(result, "Task completes successfully")
        self.assertIn(REPORT_MESSAGE, [a.get('type') for a in actions])

    @responses.activate
    def full_validate_0_5_to_2_0_conversion(self):
        setUpContextCache()
        assertion_url = 'http://example.org/assertion'
        assertion_data = {
            "uid": u"abc123™",
            "recipient": "sha256$a4a934a0bfc882a34a3e71650e40789453b2db9799a51a2d084a64caadd72397",
            "salt": "2e2bad0df9e11272ffbcee86e4c7edd4",
            "issued_on": "2017-01-01",
            "badge": {
                "name": "Test Badge for admin@bluemantis.com",
                "image": "http://example.org/image",
                "description": "Awarded using the php example codebase",
                "criteria": "http://example.org/criteria",
                "issuer": {
                    "origin": "http://example.org",
                    "name": u"Test Issuer™",
                    "org": None,
                    "contact": 'test@example.com'
                }
            }
        }

        responses.add(
            responses.GET, assertion_url,
            json=assertion_data
        )
        png_badge = os.path.join(os.path.dirname(__file__), 'testfiles', 'public_domain_heart.png')
        with open(png_badge, 'rb') as image:
            responses.add(
                responses.GET, assertion_data['badge']['image'],
                body=image.read(), status=200, content_type='image/png'
            )

        store = verification_store(assertion_url)
        state = store.get_state()
        report = generate_report(store)

        self.assertTrue(report['report']['valid'])
        assertion_node = state['graph'][0]
        badgeclass_node = state['graph'][2]
        issuer_node = state['graph'][1]

        self.assertEqual(assertion_node['id'], assertion_url)
        self.assertEqual(badgeclass_node['@context'], OPENBADGES_CONTEXT_V2_URI)
        self.assertEqual(issuer_node['type'], OBClasses.Profile)

        self.assertEqual(report['report']['openBadgesVersion'], '0.5')
